;;; kmpublic - General Functions
;;; 2007-07-30, 2021-08-11 Ver.1.2.4.20210907
;;; License: CC0

(provide "kmpublic")




;;; General

;;; 2007-07-30, 2009-10-21
(defmacro save-variables (var-list &body body)
  "Save special variables, act like save-excursion.
Example:
(progn
  (setq hide-restricted-region nil)
  (save-variables
   (*print-pretty* hide-restricted-region) ; list of variables to be saved
   ...
   (setq hide-restricted-region t)
   ...) ; end of save-variables
  hide-restricted-region)
=> nil"
  (let ((var-t (gensym)))
    `(let ((,var-t (mapcar #'eval ',var-list)))
       (unwind-protect
           (progn ,@body)
         (mapcar #'set ',var-list ,var-t)))))

(defmacro with-output-to-file ((filename &key (if-exists :overwrite) (if-does-not-exist :create)) &rest rest)
  "Redirect *standard-output* to file, like with-output-to-buffer."
  `(with-open-file (fo ,filename :direction :output :if-exists ,if-exists :if-does-not-exist ,if-does-not-exist)
     (save-variables
      (*standard-output*)
      (setq *standard-output* fo)
      ,@rest)))

(defmacro until (pred &body body)
  "(until pred ...) == (while (not pred) ...)."
  `(do ()
       (,pred)
     ,@body))

(defun append-double-quotes (filename)
  "Append double quotes to both ends, act like append-trail-slash."
  (concat "\"" (substitute-string filename "\"" "") "\""))

;;; 2011-08-25
(defmacro fsetq (symbol &body body)
  "(fsetq symbol (lambda (x1 x2 ...) body)) == (defun symbol (x1 x2 ...) body). Seealso: copy-function"
  `(setf (symbol-function ',symbol) ,@body))

;;; 2009-08-27, 2012-04-12
;;; fexpand -> pfe, Shido's lispmode.l is required because of ed::pprint-1
(when (find-symbol "pprint-1" "editor")
  (defmacro pfe (func)
    "Pritty-print function-expansions, without arguments. e.g.: (pfe apropos)"
    (let ((f (gensym)))
      `(let ((,f #',func))
         (if (and (functionp ,f) (si:*closurep ,f))
             (progn
               (save-excursion
                 (ed::pprint-1 (si:closure-body ,f)))
               (indent-sexp))
           (if (consp ,f)         ; if macro
               (progn
                 (save-excursion
                   (ed::pprint-1 ,f))
                 (indent-sexp))
             (format t "~S~%" ,f)))))))
(defun fexpand (func)
  "Expand definition of function, like macroexpand, without arguments. e.g.: (fexpand 'apropos)"
  (if (and (symbolp func) (fboundp func))
      (let ((f (funcall 'symbol-function func)))
        (if (si:*closurep f)
            (si:closure-body f)
          f))
    func))

;;; 2010-06-18, 2012-06-07
;;; overwrite pme in Shido's lispmode.l
(when (find-symbol "pprint-1" "editor")
 (defmacro pme (mac)
   "Pritty-print macro-expansions. e.g.: (pme (defun f (x) \"Double.\" (+ x x)))"
   `(progn
      (save-excursion
        (ed::pprint-1 (macroexpand-1 ',mac)))
      (indent-sexp))))

;;; How do you copy an array - comp.lang.lisp | Google Groups
;;; http://groups.google.com/group/comp.lang.lisp/browse_thread/thread/78ca9604a87c0e3d/7c50c185b5065f32?lnk=gst&q=displaced+array&rnum=9#7c50c185b5065f32
;;; Edited by Kongmu, 2009-09-05, 2020-05-24
(defun copy-array (a)
  "Copy array, like copy-seq. (equal a (copy-array a)) => t, (eq a (copy-array a)) => nil."
  (let ((adjustable (adjustable-array-p a))
        (element-type (array-element-type a))
        (fill-pointer (if (array-has-fill-pointer-p a) (fill-pointer a) nil)))
    (let ((ap (make-array (array-dimensions a)
                          :adjustable adjustable
                          :element-type element-type
                          :fill-pointer fill-pointer)))
      (dotimes (i (array-total-size a))
        (setf (row-major-aref ap i) (row-major-aref a i)))
      ap)))
(defun maparray (function array)
  "Mapping function for array. Similar as mapcar or maphash."
  (let ((array (copy-array array)))
    (dotimes (i (array-total-size array))
      (setf (row-major-aref array i) (funcall function (row-major-aref array i))))
    array))

;;; 2009-08-28, 2011-01-20
;;; ϐߑƑd] (M)
;;; http://www.shido.info/lisp/macro2.html
(defmacro with-gensyms (syms &body body)
  "Generate symbols for defmacro, not to conflict. syms is a list of symbols to be used."
  `(let ,(mapcar #'(lambda (s)
                     `(,s (gensym)))
                 syms)
     ,@body))
(defmacro for ((var start stop &optional (step 1)) &body body)
  "`for' macro in lisp. e.g. (for (i 3 6) (format t \"~D \" i)) => 3 4 5 "
  (and step (not (numberp step)) (error "step should be a number in \'for\'"))
  (and step (zerop step) (error "step should not be zero"))
  (let ((gstop (gensym)))
    `(do ((,var ,start ,(if step `(+ ,var ,step) `(1+ ,var))) (,gstop ,stop))
         (,(if (or (not step) (plusp step)) `(>= ,var ,gstop) `(<= ,var ,gstop)))
       ,@body)))

;;; from !piapro-library.l, 2010-06-18
(defmacro with-temp-directory (temp-dir &body body)
  "Work with temporary directory, act like save-excursion."
  (let ((temporary-dir (gensym)))
    `(let ((original-dir (default-directory))
           (,temporary-dir ,temp-dir))
       (cd ,temporary-dir)
       (unwind-protect
           (progn ,@body)
         (cd original-dir)))))

;;; 2010-06-18, 2011-08-25
(defmacro copy-function (from to)
  "Copy function to another symbol. Quote is not needed. e.g. (copy-function expt ^)"
  `(progn
     (setf (symbol-function ',to) #',from
           (symbol-plist ',to) (symbol-plist ',from))
     ',to))

;;; 2011-01-20
(defun make-temp-buffer-stream (name)
  "Make/clear temporary buffer and return its stream."
  (prog1
      (make-buffer-stream (switch-to-buffer name))
    (make-local-variable 'need-not-save)
    (setq need-not-save t)
    (delete-region (point-min) (point-max))))

;;; 2015-02-14, 2019-07-10
(defvar *show-name-abs-deg-output* t
  "Destination for show-name-value.")
(defvar *show-name-abs-deg-format* "~@10A = ~~{~~@7F /_ ~~@7,2F deg~~} = ~~S\n"
  "Control-string for show-name-value.")
(defmacro show-name-abs-deg (var &key (output *show-name-abs-deg-output*) (format *show-name-abs-deg-format*))
  "Show variable's name and its absolute value, phase [deg]. e.g. (let ((z #C(1 1.732))) (show-name-abs-deg z))"
  `(format ,output ,(format nil format var) (complex-to-abs-deg ,var) ,var))
(defmacro show-names-abs-deg (&rest vars)
  "Show variables' names and their absolute values, phases [deg], like show-names-values."
  (with-gensyms
   (forms)
   (if (listp (car vars))
       ; 1st argument is list, rest arguments are output and format
       (dolist (var (car vars) `(progn ,@(reverse forms)))
         (if (null (cdr vars))
             ; no rest arguments
             (push `(show-name-abs-deg ,var) forms)
           ; rest arguments are &key (output and format)
           (push `(show-name-abs-deg ,var ,@(cdr vars)) forms)))
     ; all arguments are variables
     (dolist (var vars `(progn ,@(reverse forms)))
       (push `(show-name-abs-deg ,var) forms)))))

;;; 2015-02-25, 2019-07-10
(defvar *show-name-value-output* t
  "Destination for show-name-value.")
(defvar *show-name-value-format* "~A = ~S~%"
  "Control-string for show-name-value.")
(defmacro show-name-value (var &key (output *show-name-value-output*) (format *show-name-value-format*))
  "Output variable's name and its value. e.g. (show-name-value pi :format \"~8@A = ~11,5E~%\")"
  `(format ,output ,format ',var ,var))
(defmacro show-names-values (&rest vars)
  "Output variable's name and its value. e.g. (show-names-values pi *eol-lf*)"
  (with-gensyms
   (forms)
   (if (listp (car vars))
       ; 1st argument is list, rest arguments are output and format
       (dolist (var (car vars) `(progn ,@(reverse forms)))
         (if (null (cdr vars))
             ; no rest arguments
             (push `(show-name-value ,var) forms)
           ; rest arguments are &key (output and format)
           (push `(show-name-value ,var ,@(cdr vars)) forms)))
     ; all arguments are variables
     (dolist (var vars `(progn ,@(reverse forms)))
       (push `(show-name-value ,var) forms)))))

;;; Make order of items in list random
;;; from view-random-test.l
;;; 2021-08-12 copied to kmpublic
;;; 2020-06-21, 2020-07-20
#|
(defun randomize-list-loop (l result n)
  "Internal loop for randomize-list."
  (if l                         ; 2020-07-20 null removed
      (let ((k (random n (make-random-state t))) ; 2020-06-22 (make-random-state t) added
            (a (pop l)))
        (randomize-list-loop
         l
         (append (butlast result k) (cons a (last result k)))
         (1+ n)))               ; 2020-07-04 incf -> 1+
    result))

(defun randomize-list (list)
  "Randomize items of list."
  (if list
      (let ((n 2)
            (l (copy-list list))
            result k)
        (push (pop l) result)
        (randomize-list-loop l result n))
    nil))
|#

;;; 2021-08-12 rewrited for making faster by not using recursive call
;;; 2021-08-17 refactored by using backquote
(defun randomize-list (list)
  "Randomize items of list."
  (let (result
        (original (copy-list list))
        (i 1)
        random-index)
    (while original
      (setq random-index (random i (make-random-state t)))
      (setq result `(,@(subseq result 0 random-index)
                     ,(pop original)
                     ,@(nthcdr random-index result)))
      ;(msgbox "result: ~S" result)
      (incf i))
    result))

;;; Make random index for random-vector
;;; 2020-06-21
(defun make-random-index (n)
  "Make random-ordered [0 ... n-1] list."
  (let (index)
    (dotimes (i n)
      (push i index))
    (randomize-list index)))

;;; randomize-vector
;;; 2020-06-21
(defun randomize-vector (vector)
  "Randomize items of vector."
  (let ((n (length vector)))
    (let ((result (make-vector n))
          (index (make-random-index n))
          (k 0))
      (dolist (i index result)
        (setf (svref result k) (aref vector i))
        (incf k)))))

;;; randomize-string
;;; 2020-06-22
(defun randomize-string (str)
  "Randomize characters of string."
  (let ((n (length str)))
    (let ((result (make-sequence 'string n))
          (index (make-random-index n))
          (k 0))
      (dolist (i index result)
        (setf (elt result k) (elt str i))
        (incf k)))))

;;; 2015-11-09
;;; 2021-09-07 added from .xyzzy
(defmacro vdoc (v)
  "Return documentation of variable."
  `(documentation ',v 'variable))
(defmacro fdoc (f)
  "Return documentation of function."
  `(documentation ',f 'function))

;;; 2016-12-13, 2017-10-26, 2017-11-02
;;; 2021-09-07 added from .xyzzy
(defun utf8-to-internal (input-string-or-stream &optional output-stream)
  "(convert-encoding-to-internal *encoding-utf8* str)"
  (convert-encoding-to-internal *encoding-utf8* input-string-or-stream output-stream))
;;; (utf8-to-internal "時雨")     ; => "J"
(defmacro utf8-internal (input-string-or-stream &optional output-stream)
  "(utf8-internal symbol) => (convert-encoding-to-internal *encoding-utf8* \"symbol\")"
  `(convert-encoding-to-internal *encoding-utf8* ,(string input-string-or-stream) ,output-stream))
;;; (utf8-internal 大鳳)          ; => "P"
(defun internal-to-utf8 (input-string-or-stream &optional output-stream)
  "(convert-encoding-from-internal *encoding-utf8n* str)"
  (convert-encoding-from-internal *encoding-utf8n* input-string-or-stream output-stream))
(defmacro internal-utf8 (input-string-or-stream &optional output-stream)
  "(internal-utf8 symbol) => (convert-encoding-from-internal *encoding-utf8n* \"symbol\")"
  `(convert-encoding-from-internal *encoding-utf8n* ,(string input-string-or-stream) ,output-stream))

;;; 2017-10-03
;;; 2021-09-07 added from .xyzzy
(defvar *msgbox-variables-format* "~:{~A : ~S~:^~%~}"
  "Control-string for msgbox-variables. Argument: ((name1 name2 ...) (value1 value2 ...))")
(defmacro msgbox-variables (&rest vars-list &aux (format *msgbox-variables-format*))
  "Show variables' names and their values by msgbox. e.g. (msgbox-variables pi *eol-lf*)"
  `(msgbox ,format (mapcar #'list ',vars-list (list ,@vars-list))))
; (pme (msgbox-variables pi *eol-lf*))




;;; Buffer

;;; 2008-10-27
;;; 2009-08-28 thin-out-lines-whole-buffer -> pickup-lines
(defvar *pickup-lines-n* 2
  "Default value of n for pickup-lines, make the number of lines into 1/n.")
(defun pickup-lines (&optional n)
  "Make the number of lines into 1/n."
  (interactive "*P")
  (unless (integerp n)
    (if *prefix-value* (setq n *prefix-value*)
      (setq n *pickup-lines-n*)))
  (save-excursion
    (goto-char (point-min))
    (while (next-line)
      (kill-line (1- n)))))

;;; 2009-08-28, 2013-07-03 sort-buffer-lines -> sort-lines
(defun sort-lines (&optional options)
  "Sort lines in ascending order with Windows' sort command. If you want to set options, type \"C-u M-x sort-buffer-lines\".
Type \"sort /?\" in command prompt for available options, e.g. /R reverse the result."
  (interactive "*")
  (when *prefix-args*
    (setq options (read-string "options: ")))
  (unless options (setq options ""))
  (filter-buffer (concat "sort " options)))

;;; 2009-09-17, 2013-07-03 uniq -> uniq-lines
(defun uniq-lines (&optional output-buffer)
  "Remove duplicate lines of the current buffer."
  (interactive "*")
  (let ((s (make-buffer-stream (selected-buffer)))
        line buf)
    (while (setq line (read-line s nil))
      (pushnew line buf :test #'equal))
    (when output-buffer
      (switch-to-buffer output-buffer))
    (delete-region (point-min) (point-max))
    (insert (format nil "~{~A~%~}" (reverse buf)))))

;;; 2011-10-20, 2012-01-11
(defun binary-to-string (bin-str &optional (radix 2) &aux bin)
  "\"01001001 01101110 00100000 ...\" => \"In ...\""
  (with-input-from-string
   (stream bin-str)
   (with-output-to-string
    (chr)
    (while (setq bin (read-as-string stream nil))
      (format chr "~A" (code-char (parse-integer bin :radix radix))))
    chr)))
(defun string-to-binary (str &optional (radix 2) (width 8))
  "\" ...\" -> \"10110110 10111111 10110011 ...\""
  (apply 'concat (map 'list (lambda (x) (format nil (format nil "~~~D,~D,'0R " radix width) (char-code x))) str)))

;;; 2012-06-07
(defun delete-line (&optional lines)
 "Delete lines, act like kill-line but *kill-ring* is not changed."
 (interactive "*p")
;;;  (msgbox "*prefix-value*: ~S" *prefix-value*)
 (delete-region
  (point)
  (progn
    (if (null lines)
        (if (eolp)
            (forward-line 1)
          (goto-eol))
      (if (zerop lines)
          (if (bolp)
              (forward-line -1)
            (goto-bol))
        (forward-line lines)))
    (point))))

;;; 2013-07-03, 2020-05-24
(defun reverse-lines (&optional output-buffer)
  "Reverse lines of the current buffer."
  (interactive "*")
  (let ((s (make-buffer-stream (selected-buffer)))
        line buf)
    (when (char= (char-before (point-max)) #\LFD)
      (goto-char (point-max))
      (insert "\n"))            ; read-line does not read last line consists only of \n, so add \n here
    (while (setq line (read-line s nil))
      (push line buf))
    (when output-buffer
      (switch-to-buffer output-buffer))
    (delete-region (point-min) (point-max))
    (insert (format nil "~{~A~^~%~}" buf)))) ; no \n for last line

;;; scan-lines
;;; 2021-09-07 moved from kmmisc.l
;;; 2014-06-20, 2015-03-25
(defun scan-lines (regexp &optional (input-buffer (selected-buffer)) (output-buffer "*lines*") &aux line)
  "Output lines of current buffer to *lines*, if line matched regexp."
  (interactive "sregexp: ")
  (with-input-from-buffer
   (input-buffer)
   (with-output-to-temp-buffer
    (output-buffer)
    (while (setq line (read-line *standard-input* nil))
      (when (string-match regexp line)
        (format t "~A~%" line))))))
;(global-set-key #\C-l 'scan-lines)

;;; 2015-02-13, 2015-03-25
(defun scan-lines-replace (regexp replacement &optional (input-buffer (selected-buffer)) (output-buffer "*lines*") &aux line)
  "Output lines of current buffer to *lines* with replacing, if line matched regexp."
  (interactive "sregexp: \nswith: ")
  (with-input-from-buffer
   (input-buffer)
   (with-output-to-temp-buffer
    (output-buffer)
    (while (setq line (read-line *standard-input* nil))
      (when (string-match regexp line)
        (format t "~A~%" (substitute-string line regexp replacement)))))))

;;; 2015-02-13
(defun pickup-buffer (regexp replacement &optional (input-buffer (selected-buffer)) (output-buffer "*pickup*") reverse (whole t))
  "Scan buffer by regexp pattern and output replaced strings to *pickup* with replacement."
  (interactive "sregexp: \nswith: ")
  (save-excursion
    (when whole (goto-char (point-min)))
    (let* ((regexp (compile-regexp regexp))
           (result
            (with-output-to-string
             (output-string)
;             (msgbox "point: ~A\n(scan-buffer ~A): ~A\n(match-string 0): ~A" (point) regexp (scan-buffer regexp :reverse reverse) (match-string 0))
             (while (scan-buffer regexp :reverse reverse :no-dup t)
               (format output-string "~A~%" (substitute-string (match-string 0) regexp replacement))))))
      (with-output-to-temp-buffer
       (output-buffer)
       (princ result *standard-output*))
      result)))
;;; example: for http://bible.salterrae.net/taisho/xml/matthew.xml
;;; (pickup-buffer "<w ruby=\"\\([^/<>]+?\\)\">\\([^/<>][^/<>]+?\\)</w>" "\\1,\\2")




;;; Interactive

(defun replace-regexp-whole (regexp replacement &optional case-fold)
  "replace-regexp for whole buffer."
  (interactive "*sReplace regexp: \nswith: ")
  (save-excursion
    (goto-char (point-min))
    (replace-regexp (compile-regexp regexp case-fold) replacement t)))

(defun delete-and-insert (&optional (string ""))
  "Delete selection and insert string."
  (when (get-selection-type) (delete-char-or-selection))
  (insert string))

(defun idate ()
  "Insert %Y-%m-%d type date."
  (interactive "*")
  (delete-and-insert (format-date-string "%Y-%m-%d")))

(defun udate ()
  "Insert %YN%#m%#d type date."
  (interactive "*")
  (delete-and-insert (format-date-string "%YN%#m%#d")))

(defun vdate ()
  "Insert selected type date."
  (interactive "*")
  (delete-and-insert)
  (save-variables
   (*date-formats*)
   (setq *date-formats*
         '("%Y-%m-%d"
           "%Y-%m-%dT%H:%M:%S%#Z"
           "%YN%#m%#d"
           "%YN%#m%#d(%v)"
           "%g%#eN%#m%#d"
           "%g%#eN%#m%#d %vj"
           " %H:%M:%S"
           " %#H:%M:%S"
           " %#H%#M%#Sb"
           " %p%#I%#M%#Sb"))
   (insert-date-string)))

;;; ?, 2021-09-07
(defun open-file-with-windows ()
  "Open current file with Windows file associations."
  (interactive)
  (let ((file (get-buffer-file-name)))
    (if file (shell-execute file)
      (let ((file (merge-pathnames (buffer-name (selected-buffer)) (default-directory)))) ; for xdoc2txt
        (if (file-exist-p file) (shell-execute file)
          (error "Current buffer isn't a file."))))))

;;; 2009-09-03
(defun toggle-narrowing ()
  "Toggle narrow-to-region and widen."
  (interactive)
  (if (and (= (point-min) 0)
           (= (point-max) (buffer-size)))
      (values t (call-interactively 'narrow-to-region))
    (values nil (widen))))

;;; 2010-04-04
(defun kill-window ()
  "kill-selected-buffer and delete-window."
  (interactive)
  (kill-selected-buffer)
  (delete-window))

(defun md5-file (path)
  "Output MD5 from file."
  (interactive "fFile: ")
  (with-open-file
   (stream path)
   (message (si:md5 stream))
   (si:md5 stream)))




;;; WWW

;;; 2007-08-05, 2009-05-20
(defvar *uri-to-filename-no-ext* nil
  "If t, extension will be omitted.")

(defun uri-to-filename-str (uri)
  "Convert / into ; and escape non-ASCII charactors."
  (substitute-string (si:www-url-encode (substitute-string uri "^\\(https?\\|ftp\\|gopher\\|urn\\)://" "") nil "0-9A-Za-z---._~/+") "/" ";"))

(defun uri-to-filename (&optional uri (no-ext *uri-to-filename-no-ext*) &aux filename)
  "Convert / into ; and escape non-ASCII charactors. Result will be put into clipboard."
  (interactive)
  (when (prog1 (unless uri (setq uri (get-clipboard-data)))
          (setq filename (uri-to-filename-str uri)))
    (when no-ext (setq filename (substitute-string filename "\\.[^;]+?$" "")))
    (copy-to-clipboard filename)
    (message "~A" filename))
  filename)

;;; 2011-10-20
(defun www-url-encode-utf8 (input-string-or-stream &optional output-stream literal-chars)
  "Encode URL in UTF-8."
  (si:www-url-encode (map-internal-to-utf-8 input-string-or-stream) output-stream literal-chars))
(defun www-url-decode-utf8 (input-string-or-stream &optional output-stream)
  "Decode URL as UTF-8."
  (or (format output-stream "~A" (map-utf-8-to-internal (si:www-url-decode input-string-or-stream nil))) output-stream))
